#include "ELMduino.h"


bool ELM327::begin(Stream &stream, char protocol, uint16_t payloadLen)
{
	elm_port = &stream;
	PAYLOAD_LEN = payloadLen;

	payload = (char*)malloc(PAYLOAD_LEN);

	// test if serial port is connected
	if (!elm_port)
		return false;

	// try to connect
	if (!initializeELM(protocol))
		return false;
  
	return true;
}

//____________________________________________________________

bool ELM327::initializeELM(char protocol)
{
	char command[10] = { '\0' };
	connected = false;
	
	sendCommand(RESET_ALL);
        delay(100);

	sendCommand(ECHO_OFF);
	delay(100);

	sendCommand(PRINTING_SPACES_OFF);
	delay(100);
	
	sendCommand(ALLOW_LONG_MESSAGES);
	delay(100);

	// Set protocol
	sprintf(command, TRY_PROT_H_AUTO_SEARCH, protocol);

	if (sendCommand(command) == ELM_SUCCESS)
	{
		if (strstr(payload, "OK") != NULL)
		{
			connected = true;
			return connected;
		}
	}
	
	// Set protocol and save
	sprintf(command, SET_PROTOCOL_TO_H_SAVE, protocol);

    if (sendCommand(command) == ELM_SUCCESS)
        if (strstr(payload, "OK") != NULL)
            connected = true;

	return connected;
}

//____________________________________________________________

void ELM327::formatQueryArray(uint8_t service, uint16_t pid)
{
	query[0] = ((service >> 4) & 0xF) + '0';
	query[1] = (service & 0xF) + '0';

	// determine PID length (standard queries have 16-bit PIDs,
	// but some custom queries have PIDs with 32-bit values)
	if (pid & 0xFF00)
	{
		longQuery = true;

		query[2] = ((pid >> 12) & 0xF) + '0';
		query[3] = ((pid >> 8) & 0xF) + '0';
		query[4] = ((pid >> 4) & 0xF) + '0';
		query[5] = (pid & 0xF) + '0';

		upper(query, 6);
	}
	else
	{
		longQuery = false;

		query[2] = ((pid >> 4) & 0xF) + '0';
		query[3] = (pid & 0xF) + '0';
		query[4] = '\0';
		query[5] = '\0';

		upper(query, 4);
	}
}

//____________________________________________________________

void ELM327::upper(char string[], uint8_t buflen)
{
	for (uint8_t i = 0; i < buflen; i++)
	{
		if (string[i] > 'Z')
			string[i] -= 32;
		else if ((string[i] > '9') && (string[i] < 'A'))
			string[i] += 7;
	}
}

//____________________________________________________________

bool ELM327::timeout()
{
	currentTime = millis();
	if ((currentTime - previousTime) >= timeout_ms)
		return true;
	return false;
}

//____________________________________________________________

uint8_t ELM327::ctoi(uint8_t value)
{
	if (value >= 'A')
		return value - 'A' + 10;
	else
		return value - '0';
}

//____________________________________________________________

int8_t ELM327::nextIndex(char const *str,
                         char const *target,
                         uint8_t numOccur=1)
{
	char const *p = str;
	char const *r = str;
	uint8_t count;

	for (count = 0; ; ++count)
	{
		p = strstr(p, target);

		if (count == (numOccur - 1))
			break;

		if (!p)
			break;

		p++;
	}

	if (!p)
		return -1;

	return p - r;
}

//____________________________________________________________

void ELM327::flushInputBuff()
{
	while (elm_port->available())
		elm_port->read();
}

//____________________________________________________________

bool ELM327::queryPID(uint8_t service,
                      uint16_t pid)
{
	if (connected)
	{
		formatQueryArray(service, pid);
		sendCommand(query);

		return true;
	}
	
	return false;
}

//____________________________________________________________

bool ELM327::queryPID(char queryStr[])
{
	if (connected)
	{
		if (strlen(queryStr) <= 4)
			longQuery = false;
		else
			longQuery = true;

		sendCommand(queryStr);

		return true;
	}

	return false;
}


//____________________________________________________________

float ELM327::Coolant()
{
	if (queryPID(SERVICE_01, ENGINE_COOLANT_TEMP))
		return (findResponse() - 40.0);

	return ELM_GENERAL_ERROR;
}

//____________________________________________________________

int32_t ELM327::speedo()
{
	if (queryPID(SERVICE_01, VEHICLE_SPEED))
		return findResponse();

	return ELM_GENERAL_ERROR;
}

//____________________________________________________________

//float ELM327::mph()
//{
//	float mph = speed();
//
//	if (status == ELM_SUCCESS)
//		return mph * KPH_MPH_CONVERT;
//
//	return ELM_GENERAL_ERROR;
//}

//____________________________________________________________

float ELM327::Rpm()
{
  if (queryPID(SERVICE_01, ENGINE_RPM))
    return (findResponse() / 4.0);

  return ELM_GENERAL_ERROR;
}

//____________________________________________________________

float ELM327::volt()
{
	if (queryPID(SERVICE_01, CONTROL_MODULE_VOLTAGE))
		return findResponse();

	return ELM_GENERAL_ERROR;
}

//____________________________________________________________

float ELM327::driving()
{
	if (queryPID(SERVICE_01, RUN_TIME_SINCE_ENGINE_START))
		return (findResponse() / 60);

	return ELM_GENERAL_ERROR;
}

//____________________________________________________________

float ELM327::Intake()
{
	if (queryPID(SERVICE_01, INTAKE_AIR_TEMP))
		return (findResponse() - 40.0);

	return ELM_GENERAL_ERROR;
}

//____________________________________________________________

float ELM327::Ambient()
{
	if (queryPID(SERVICE_01, AMBIENT_AIR_TEMP))
		return (findResponse() - 40.0);

	return ELM_GENERAL_ERROR;
}

//____________________________________________________________

int8_t ELM327::sendCommand(const char *cmd)
{
	uint8_t counter = 0;
	connected = false;

	for (byte i = 0; i < PAYLOAD_LEN; i++)
		payload[i] = '\0';

	// reset input buffer and number of received bytes
	recBytes = 0;
	flushInputBuff();

	elm_port->print(cmd);
	elm_port->print('\r');

	// prime the timeout timer
	previousTime = millis();
	currentTime  = previousTime;

	// buffer the response of the ELM327 until either the
	// end marker is read or a timeout has occurred
	while ((counter < (PAYLOAD_LEN + 1)) && !timeout())
	{
		if (elm_port->available())
		{
			char recChar = elm_port->read();

			if (recChar == '>')
				break;
			else if (isspace(recChar))
				continue;
			
			payload[counter] = recChar;
			counter++;
		}
	}

	if (timeout())
	{
		status = ELM_TIMEOUT;
		return status;
	}
	
	if (nextIndex(payload, "UNABLETOCONNECT") >= 0)
	{
		status = ELM_UNABLE_TO_CONNECT;
		return status;
	}

	connected = true;

	if (nextIndex(payload, "NODATA") >= 0)
	{
		status = ELM_NO_DATA;
		return status;
	}

	if (nextIndex(payload, "STOPPED") >= 0)
	{
		status = ELM_STOPPED;
		return status;
	}

	if (nextIndex(payload, "ERROR") >= 0)
	{
		status = ELM_GENERAL_ERROR;
		return status;
	}

	// keep track of how many bytes were received in
	// the ELM327's response (not counting the
	// end-marker '>') if a valid response is found
	recBytes = counter;

	status = ELM_SUCCESS;
	return status;
}




/*
 uint64_t ELM327::findResponse()

 Description:
 ------------
  * Parses the buffered ELM327's response and returns the queried data

 Inputs:
 -------
  * void

 Return:
 -------
  * uint64_t - Query response value
*/
uint64_t ELM327::findResponse()
{
	uint8_t firstDatum = 0;
	uint8_t payBytes = 0;
	char header[7] = { '\0' };

	if (longQuery)
	{
		header[0] = query[0] + 4;
		header[1] = query[1];
		header[2] = query[2];
		header[3] = query[3];
		header[4] = query[4];
		header[5] = query[5];
	}
	else
	{
		header[0] = query[0] + 4;
		header[1] = query[1];
		header[2] = query[2];
		header[3] = query[3];
	}

	int8_t firstHeadIndex  = nextIndex(payload, header);
	int8_t secondHeadIndex = nextIndex(payload, header, 2);

	if (firstHeadIndex >= 0)
	{
		if (longQuery)
			firstDatum = firstHeadIndex + 6;
		else
			firstDatum = firstHeadIndex + 4;

		// Some ELM327s (such as my own) respond with two
		// "responses" per query. "payBytes" represents the
		// correct number of bytes returned by the ELM327
		// regardless of how many "responses" were returned
		if (secondHeadIndex >= 0)
			payBytes = secondHeadIndex - firstDatum;
		else
			payBytes = recBytes - firstDatum;


		uint64_t response = 0;
		for(uint8_t i = 0; i < payBytes; i++)
		{
			uint8_t payloadIndex = firstDatum + i;
			uint8_t bitsOffset = 4 * (payBytes - i - 1);
			response = response | (ctoi(payload[payloadIndex]) << bitsOffset);
		}

		// It is usefull to have the response bytes
		// broken-out because some PID algorithms (standard
		// and custom) require special operations for each
		// byte returned
		responseByte_0 = response & 0xFF;
		responseByte_1 = (response >> 8) & 0xFF;
		responseByte_2 = (response >> 16) & 0xFF;
		responseByte_3 = (response >> 24) & 0xFF;
		responseByte_4 = (response >> 32) & 0xFF;
		responseByte_5 = (response >> 40) & 0xFF;
		responseByte_6 = (response >> 48) & 0xFF;
		responseByte_7 = (response >> 56) & 0xFF;
		
		return response;
	}

	return 0;
}
